@@ -1,7 +1,7 @@
 /*
  *   fs/cifs/cifsfs.c
  *
- *   Copyright (c) International Business Machines  Corp., 2002
+ *   Copyright (C) International Business Machines  Corp., 2002,2003
  *   Author(s): Steve French (sfrench@us.ibm.com)
  *
  *   Common Internet FileSystem (CIFS) client
@@ -30,7 +30,9 @@
 #include <linux/init.h>
 #include <linux/version.h>
 #include <linux/list.h>
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 #include <linux/seq_file.h>
+#endif
 #include <linux/vfs.h>
 #include "cifsfs.h"
 #include "cifspdu.h"
@@ -42,12 +44,18 @@
 #include <linux/mm.h>
 #define CIFS_MAGIC_NUMBER 0xFF534D42	/* the first four bytes of all SMB PDUs */
 
+#ifdef CIFS_QUOTA
+static struct quotactl_ops cifs_quotactl_ops;
+#endif
+
 extern struct file_system_type cifs_fs_type;
 
 int cifsFYI = 0;
 int cifsERROR = 1;
 int traceSMB = 0;
 unsigned int oplockEnabled = 1;
+unsigned int quotaEnabled = 0;
+unsigned int linuxExtEnabled = 1;
 unsigned int lookupCacheEnabled = 1;
 unsigned int multiuser_mount = 0;
 unsigned int extended_security = 0;
@@ -75,7 +83,6 @@ cifs_read_super(struct super_block *sb, void *data, int silent)
 	cifs_sb = CIFS_SB(sb);
 	if(cifs_sb == NULL)
 		return 0;
-	cifs_sb->local_nls = load_nls_default();	/* needed for ASCII cp to Unicode converts */
 
 	rc = cifs_mount(sb, cifs_sb, data, NULL);
 
@@ -90,7 +97,9 @@ cifs_read_super(struct super_block *sb, void *data, int silent)
 	sb->s_op = &cifs_super_ops;
 /*	if(cifs_sb->tcon->ses->server->maxBuf > MAX_CIFS_HDR_SIZE + 512)
 	    sb->s_blocksize = cifs_sb->tcon->ses->server->maxBuf - MAX_CIFS_HDR_SIZE; */
-
+#ifdef CIFS_QUOTA
+	sb->s_qcop = &cifs_quotactl_ops;
+#endif
 	sb->s_blocksize = CIFS_MAX_MSGSIZE;
 	sb->s_blocksize_bits = 14;	/* default 2**14 = CIFS_MAX_MSGSIZE */
 	inode = iget(sb, ROOT_I);
@@ -214,6 +223,7 @@ cifs_destroy_inode(struct inode *inode)
  * Not all settable options are displayed but most of the important
  * ones are.
  */
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 static int
 cifs_show_options(struct seq_file *s, struct vfsmount *m)
 {
@@ -236,6 +246,112 @@ cifs_show_options(struct seq_file *s, struct vfsmount *m)
 	}
 	return 0;
 }
+#endif
+
+#ifdef CIFS_QUOTA
+int cifs_xquota_set(struct super_block * sb, int quota_type, qid_t qid,
+                struct fs_disk_quota * pdquota)
+{
+        int xid;
+        int rc = 0;
+        struct cifs_sb_info *cifs_sb = CIFS_SB(sb);
+        struct cifsTconInfo *pTcon;
+
+        if(cifs_sb)
+                pTcon = cifs_sb->tcon;
+        else
+                return -EIO;
+
+
+        xid = GetXid();
+        if(pTcon) {
+                cFYI(1,("set type: 0x%x id: %d",quota_type,qid));
+        } else {
+                return -EIO;
+        }
+
+        FreeXid(xid);
+        return rc;
+}
+
+int cifs_xquota_get(struct super_block * sb, int quota_type, qid_t qid,
+                struct fs_disk_quota * pdquota)
+{
+        int xid;
+        int rc = 0;
+        struct cifs_sb_info *cifs_sb = CIFS_SB(sb);
+        struct cifsTconInfo *pTcon;
+
+        if(cifs_sb)
+                pTcon = cifs_sb->tcon;
+        else
+                return -EIO;
+
+        xid = GetXid();
+        if(pTcon) {
+                cFYI(1,("set type: 0x%x id: %d",quota_type,qid));
+        } else {
+                rc = -EIO;
+        }
+
+        FreeXid(xid);
+        return rc;
+}
+
+int cifs_xstate_set(struct super_block * sb, unsigned int flags, int operation)
+{
+        int xid;
+        int rc = 0;
+        struct cifs_sb_info *cifs_sb = CIFS_SB(sb);
+        struct cifsTconInfo *pTcon;
+
+        if(cifs_sb)
+                pTcon = cifs_sb->tcon;
+        else
+                return -EIO;
+
+        xid = GetXid();
+        if(pTcon) {
+                cFYI(1,("flags: 0x%x operation: 0x%x",flags,operation));
+        } else {
+                rc = -EIO;
+        }
+
+        FreeXid(xid);
+        return rc;
+}
+
+int cifs_xstate_get(struct super_block * sb, struct fs_quota_stat *qstats)
+{
+        int xid;
+        int rc = 0;
+        struct cifs_sb_info *cifs_sb = CIFS_SB(sb);
+        struct cifsTconInfo *pTcon;
+
+        if(cifs_sb) {
+                pTcon = cifs_sb->tcon;
+        } else {
+                return -EIO;
+        }
+        xid = GetXid();
+        if(pTcon) {
+                cFYI(1,("pqstats %p",qstats));
+        } else {
+                rc = -EIO;
+        }
+
+        FreeXid(xid);
+        return rc;
+}
+
+static struct quotactl_ops cifs_quotactl_ops = {
+        .set_xquota     = cifs_xquota_set,
+        .get_xquota     = cifs_xquota_set,
+        .set_xstate     = cifs_xstate_set,
+        .get_xstate     = cifs_xstate_get,
+};
+#endif
+
 
 struct super_operations cifs_super_ops = {
 	.read_inode = cifs_read_inode,
@@ -247,7 +363,7 @@ struct super_operations cifs_super_ops = {
 	.delete_inode	= cifs_delete_inode,  *//* Do not need the above two functions     
    unless later we add lazy close of inodes or unless the kernel forgets to call
    us with the same number of releases (closes) as opens */
-#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,17))
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 	.show_options = cifs_show_options,
 #endif
 /*    .umount_begin   = cifs_umount_begin, *//* consider adding in the future */
@@ -394,6 +510,9 @@ cifs_init_inodecache(void)
 					      cifs_init_once, NULL);
 	if (cifs_inode_cachep == NULL)
 		return -ENOMEM;
+	else
+		memset(cifs_sb,0,sizeof(struct cifs_sb_info));
+	
 
 	return 0;
 }
@@ -458,8 +577,6 @@ cifs_destroy_mids(void)
 
 static int cifs_oplock_thread(void * dummyarg)
 {
-	struct list_head * tmp;
-	struct list_head * tmp1;
 	struct oplock_q_entry * oplock_item;
 	struct cifsTconInfo *pTcon;
 	struct inode * inode;
@@ -471,37 +588,43 @@ static int cifs_oplock_thread(void * dummyarg)
 	oplockThread = current;
 	do {
 		set_current_state(TASK_INTERRUPTIBLE);
-		schedule_timeout(100*HZ);
-		/* BB add missing code */
-		write_lock(&GlobalMid_Lock); 
-		list_for_each_safe(tmp, tmp1, &GlobalOplock_Q) {
-			oplock_item = list_entry(tmp, struct oplock_q_entry,
-							       qhead);
+
+		schedule_timeout(3*HZ);
+		spin_lock(&GlobalMid_Lock);
+		if(list_empty(&GlobalOplock_Q)) {
+			spin_unlock(&GlobalMid_Lock);
+			schedule_timeout(7*HZ);
+		} else {
+			oplock_item = list_entry(GlobalOplock_Q.next,
+				struct oplock_q_entry, qhead);
+
 			if(oplock_item) {
 				pTcon = oplock_item->tcon;
 				inode = oplock_item->pinode;
 				netfid = oplock_item->netfid;
+				spin_unlock(&GlobalMid_Lock);
 				DeleteOplockQEntry(oplock_item);
-				write_unlock(&GlobalMid_Lock);
+				if (S_ISREG(inode->i_mode)) {
 #if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,17))
-				rc = filemap_fdatasync(inode->i_mapping);
-				if(rc)
-					CIFS_I(inode)->write_behind_rc 
-						= rc;
+					rc = filemap_fdatasync(inode->i_mapping);
+					if(rc)
+						CIFS_I(inode)->write_behind_rc = rc;
 #else 
-				filemap_fdatasync(pinode->i_mapping);
+					filemap_fdatasync(inode->i_mapping);
 #endif
+					if(CIFS_I(inode)->clientCanCacheRead == 0) 
+						invalidate_inode_pages(inode);
+				} else
+					rc = 0;
 				rc = CIFSSMBLock(0, pTcon, 
 					netfid,
 					0 /* len */ , 0 /* offset */, 0, 
 					0, LOCKING_ANDX_OPLOCK_RELEASE,
 					0 /* wait flag */);
 				cFYI(1,("Oplock release rc = %d ",rc));
-				write_lock(&GlobalMid_Lock);
 			} else
-				break;
+				spin_unlock(&GlobalMid_Lock);
 		}
-		write_unlock(&GlobalMid_Lock);
 	} while(!signal_pending(current));
 	complete_and_exit (&cifs_oplock_exited, 0);
 }
@@ -522,13 +645,16 @@ init_cifs(void)
  */
 	atomic_set(&sesInfoAllocCount, 0);
 	atomic_set(&tconInfoAllocCount, 0);
+	atomic_set(&tcpSesReconnectCount, 0);
+	atomic_set(&tconInfoReconnectCount, 0);
+
 	atomic_set(&bufAllocCount, 0);
 	atomic_set(&midCount, 0);
 	GlobalCurrentXid = 0;
 	GlobalTotalActiveXid = 0;
 	GlobalMaxActiveXid = 0;
 	GlobalSMBSeslock = RW_LOCK_UNLOCKED;
-	GlobalMid_Lock = RW_LOCK_UNLOCKED;
+	GlobalMid_Lock = SPIN_LOCK_UNLOCKED;
 
 /*	rc = cifs_init_inodecache();*/
 	if (!rc) {
