@@ -1,7 +1,7 @@
 /*
  *   fs/cifs/cifssmb.c
  *
- *   Copyright (c) International Business Machines  Corp., 2002
+ *   Copyright (C) International Business Machines  Corp., 2002,2003
  *   Author(s): Steve French (sfrench@us.ibm.com)
  *
  *   Contains the routines for constructing the SMB PDUs themselves
@@ -21,11 +21,16 @@
  *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */
 
- /* SMB/CIFS PDU handling routines here - except for leftovers in connect.c */
+ /* SMB/CIFS PDU handling routines here - except for leftovers in connect.c   */
+ /* These are mostly routines that operate on a pathname, or on a tree id     */
+ /* (mounted volume), but there are eight handle based routines which must be */
+ /* treated slightly different for reconnection purposes since we never want  */
+ /* to reuse a stale file handle and the caller knows the file handle */
 
 #include <linux/fs.h>
 #include <linux/kernel.h>
 #include <linux/vfs.h>
+#include <linux/version.h>
 #include <asm/uaccess.h>
 #include "cifspdu.h"
 #include "cifsglob.h"
@@ -37,31 +42,104 @@ static struct {
 	int index;
 	char *name;
 } protocols[] = {
-	{
-	CIFS_PROT, "\2NT LM 0.12"}, {
-	BAD_PROT, "\2"}
+	{CIFS_PROT, "\2NT LM 0.12"}, 
+	{BAD_PROT, "\2"}
 };
 
+
+/* Mark as invalid, all open files on tree connections since they
+   were closed when session to server was lost */
+void mark_open_files_invalid(struct cifsTconInfo * pTcon)
+{
+	struct cifsFileInfo *open_file = NULL;
+	struct list_head * tmp;
+	struct list_head * tmp1;
+
+/* list all files open on tree connection and mark them invalid */
+	write_lock(&GlobalSMBSeslock);
+	list_for_each_safe(tmp, tmp1, &pTcon->openFileList) {
+		open_file = list_entry(tmp,struct cifsFileInfo, tlist);
+		if(open_file) {
+			open_file->invalidHandle = TRUE;
+		}
+	}
+	write_unlock(&GlobalSMBSeslock);
+	/* BB Add call to invalidate_inodes(sb) for all superblocks mounted to this tcon */
+}
+
 int
 smb_init(int smb_command, int wct, struct cifsTconInfo *tcon,
 	 void **request_buf /* returned */ ,
 	 void **response_buf /* returned */ )
 {
+	int timeout = 10 * HZ;
 	int rc = 0;
 
-	if(tcon && (tcon->tidStatus == CifsNeedReconnect)) {
-		rc = -EIO;
-		if(tcon->ses) {
-			struct nls_table *nls_codepage = load_nls_default();
+	/* SMBs NegProt, SessSetup, uLogoff do not have tcon yet so
+	   check for tcp and smb session status done differently
+	   for those three - in the calling routine */
+	if(tcon) {
+		if((tcon->ses) && (tcon->ses->server)){
+			struct nls_table *nls_codepage;
+				/* Give Demultiplex thread up to 10 seconds to 
+					reconnect, should be greater than cifs socket
+					timeout which is 7 seconds */
+			while(tcon->ses->server->tcpStatus == CifsNeedReconnect) {
+		                while ((tcon->ses->server->tcpStatus != CifsGood) && (timeout > 0)){
+					timeout = interruptible_sleep_on_timeout(&tcon->ses->server->response_q,timeout);
+				}
+				if(tcon->ses->server->tcpStatus == CifsNeedReconnect) {
+				/* on "soft" mounts we wait once */
+					if((tcon->retry == FALSE) ||
+					   (tcon->ses->status == CifsExiting)) {
+						cFYI(1,("gave up waiting on reconnect in smb_init"));
+						return -EHOSTDOWN;
+					} /* else "hard" mount - keep retrying until
+					process is killed or server comes back up */
+				} else /* TCP session is reestablished now */
+					break;
+
+			}
+			
+			nls_codepage = load_nls_default();
+		/* need to prevent multiple threads trying to
+		simultaneously reconnect the same SMB session */
+			down(&tcon->ses->sesSem);
 			if(tcon->ses->status == CifsNeedReconnect)
 				rc = setup_session(0, tcon->ses, nls_codepage);
-			if(!rc) {
+			if(!rc && (tcon->tidStatus == CifsNeedReconnect)) {
+				mark_open_files_invalid(tcon);
 				rc = CIFSTCon(0, tcon->ses, tcon->treeName, tcon,
 					nls_codepage);
+				up(&tcon->ses->sesSem);
+				if(rc == 0)
+					atomic_inc(&tconInfoReconnectCount);
+
 				cFYI(1, ("reconnect tcon rc = %d", rc));
-				if(!rc)
-					reopen_files(tcon,nls_codepage);
+				/* Removed call to reopen open files here - 
+					it is safer (and faster) to reopen files
+					one at a time as needed in read and write */
+
+				/* Check if handle based operation so we 
+					know whether we can continue or not without
+					returning to caller to reset file handle */
+				switch(smb_command) {
+					case SMB_COM_READ_ANDX:
+					case SMB_COM_WRITE_ANDX:
+					case SMB_COM_CLOSE:
+					case SMB_COM_FIND_CLOSE2:
+					case SMB_COM_LOCKING_ANDX: {
+						unload_nls(nls_codepage);
+						return -EAGAIN;
+					}
+				}
+			} else {
+				up(&tcon->ses->sesSem);
 			}
+			unload_nls(nls_codepage);
+
+		} else {
+			return -EIO;
 		}
 	}
 	if(rc)
@@ -97,7 +175,6 @@ CIFSSMBNegotiate(unsigned int xid, struct cifsSesInfo *ses)
 		rc = -EIO;
 		return rc;
 	}
-
 	rc = smb_init(SMB_COM_NEGOTIATE, 0, 0 /* no tcon yet */ ,
 		      (void **) &pSMB, (void **) &pSMBr);
 	if (rc)
@@ -217,12 +294,19 @@ CIFSSMBTDis(const int xid, struct cifsTconInfo *tcon)
 		return -EBUSY;
 	}
 
+	/* No need to return error on this operation if tid invalidated and 
+	closed on server already e.g. due to tcp session crashing */
+	if(tcon->tidStatus == CifsNeedReconnect) {
+		up(&tcon->tconSem);
+		return 0;  
+	}
+
 /* BB remove (from server) list of shares - but with smp safety  BB */
 /* BB is ses active - do we need to check here - but how? BB */
-    if((tcon->ses == 0) || (tcon->ses->server == 0)) {    
-        up(&tcon->tconSem);
-        return -EIO;
-    }
+	if((tcon->ses == 0) || (tcon->ses->server == 0)) {    
+		up(&tcon->tconSem);
+		return -EIO;
+	}
 
 	rc = smb_init(SMB_COM_TREE_DISCONNECT, 0, tcon,
 		      (void **) &smb_buffer, (void **) &smb_buffer_response);
@@ -238,6 +322,12 @@ CIFSSMBTDis(const int xid, struct cifsTconInfo *tcon)
 	if (smb_buffer)
 		buf_release(smb_buffer);
 	up(&tcon->tconSem);
+
+	/* No need to return error on this operation if tid invalidated and 
+	closed on server already e.g. due to tcp session crashing */
+	if (rc == -EAGAIN)
+		rc = 0;
+
 	return rc;
 }
 
@@ -250,9 +340,8 @@ CIFSSMBLogoff(const int xid, struct cifsSesInfo *ses)
 	int length;
 
 	cFYI(1, ("In SMBLogoff for session disconnect"));
-
 	if (ses)
-		down(&ses->sesSem); /* check this sem more places */
+		down(&ses->sesSem);
 	else
 		return -EIO;
 
@@ -286,6 +375,12 @@ CIFSSMBLogoff(const int xid, struct cifsSesInfo *ses)
 	if (pSMB)
 		buf_release(pSMB);
 	up(&ses->sesSem);
+
+	/* if session dead then we do not need to do ulogoff,
+		since server closed smb session, no sense reporting 
+		error */
+	if (rc == -EAGAIN)
+		rc = 0;
 	return rc;
 }
 
@@ -299,6 +394,7 @@ CIFSSMBDelFile(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned;
 	int name_len;
 
+DelFileRetry:
 	rc = smb_init(SMB_COM_DELETE, 1, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -329,6 +425,9 @@ CIFSSMBDelFile(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto DelFileRetry;
+
 	return rc;
 }
 
@@ -343,7 +442,7 @@ CIFSSMBRmDir(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In CIFSSMBRmDir"));
-
+RmDirRetry:
 	rc = smb_init(SMB_COM_DELETE_DIRECTORY, 0, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -372,6 +471,8 @@ CIFSSMBRmDir(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto RmDirRetry;
 	return rc;
 }
 
@@ -386,7 +487,7 @@ CIFSSMBMkDir(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In CIFSSMBMkDir"));
-
+MkDirRetry:
 	rc = smb_init(SMB_COM_CREATE_DIRECTORY, 0, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -415,7 +516,8 @@ CIFSSMBMkDir(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
-
+	if (rc == -EAGAIN)
+		goto MkDirRetry;
 	return rc;
 }
 
@@ -432,6 +534,7 @@ CIFSSMBOpen(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned;
 	int name_len;
 
+openRetry:
 	rc = smb_init(SMB_COM_NT_CREATE_ANDX, 24, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -501,7 +604,8 @@ CIFSSMBOpen(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
-
+	if (rc == -EAGAIN)
+		goto openRetry;
 	return rc;
 }
 
@@ -526,6 +630,10 @@ CIFSSMBRead(const int xid, struct cifsTconInfo *tcon,
 	if (rc)
 		return rc;
 
+        /* tcon and ses pointer are checked in smb_init */
+        if (tcon->ses->server == NULL)
+                return -ECONNABORTED;
+
 	pSMB->AndXCommand = 0xFF;	/* none */
 	pSMB->Fid = netfid;
 	pSMB->OffsetLow = cpu_to_le32(lseek & 0xFFFFFFFF);
@@ -566,6 +674,9 @@ CIFSSMBRead(const int xid, struct cifsTconInfo *tcon,
 		else
 			*buf = (char *)pSMB;
 	}
+
+	/* Note: On -EAGAIN error only caller can retry on handle based calls 
+		since file handle passed in no longer valid */
 	return rc;
 }
 
@@ -584,6 +695,9 @@ CIFSSMBWrite(const int xid, struct cifsTconInfo *tcon,
 		      (void **) &pSMBr);
 	if (rc)
 		return rc;
+	/* tcon and ses pointer are checked in smb_init */
+	if (tcon->ses->server == NULL)
+		return -ECONNABORTED;
 
 	pSMB->AndXCommand = 0xFF;	/* none */
 	pSMB->Fid = netfid;
@@ -617,6 +731,9 @@ CIFSSMBWrite(const int xid, struct cifsTconInfo *tcon,
 	if (pSMB)
 		buf_release(pSMB);
 
+	/* Note: On -EAGAIN error only caller can retry on handle based calls 
+		since file handle passed in no longer valid */
+
 	return rc;
 }
 
@@ -626,45 +743,59 @@ CIFSSMBLock(const int xid, struct cifsTconInfo *tcon,
 	    const __u64 offset, const __u32 numUnlock,
 	    const __u32 numLock, const __u8 lockType, const int waitFlag)
 {
-	int rc = 0;
-	LOCK_REQ *pSMB = NULL;
-	LOCK_RSP *pSMBr = NULL;
-	int bytes_returned;
-	int timeout = 0;
+		int rc = 0;
+		LOCK_REQ *pSMB = NULL;
+		LOCK_RSP *pSMBr = NULL;
+		int bytes_returned;
+		int timeout = 0;
+		__u64 temp;
 
-	cFYI(1, ("In CIFSSMBLock"));
+		cFYI(1, ("In CIFSSMBLock"));
+		rc = smb_init(SMB_COM_LOCKING_ANDX, 8, tcon, (void **) &pSMB,
+				  (void **) &pSMBr);
+		if (rc)
+			return rc;
+
+		if(lockType == LOCKING_ANDX_OPLOCK_RELEASE) {
+			timeout = -1; /* no response expected */
+			pSMB->Timeout = 0;
+		} else if (waitFlag == TRUE) {
+			timeout = 3; /* blocking operation, no timeout */
+			pSMB->Timeout = -1; /* blocking - do not time out */
+		} else {
+			pSMB->Timeout = 0;
+		}
+
+		pSMB->NumberOfLocks = cpu_to_le32(numLock);
+		pSMB->NumberOfUnlocks = cpu_to_le32(numUnlock);
+		pSMB->LockType = lockType;
+		pSMB->AndXCommand = 0xFF;	/* none */
+		pSMB->Fid = smb_file_id; /* netfid stays le */
+
+		pSMB->Locks[0].Pid = cpu_to_le16(current->tgid);
+		temp = cpu_to_le64(len);
+		pSMB->Locks[0].LengthLow = (__u32)(len & 0xFFFFFFFF);
+		pSMB->Locks[0].LengthHigh =  (__u32)(len>>32);
+		temp = cpu_to_le64(offset);
+		pSMB->Locks[0].OffsetLow = (__u32)(offset & 0xFFFFFFFF);
+		pSMB->Locks[0].OffsetHigh = (__u32)(offset>>32);
+		pSMB->ByteCount = sizeof (LOCKING_ANDX_RANGE);
+		pSMB->hdr.smb_buf_length += pSMB->ByteCount;
+		pSMB->ByteCount = cpu_to_le16(pSMB->ByteCount);
+
+		rc = SendReceive(xid, tcon->ses, (struct smb_hdr *) pSMB,
+				 (struct smb_hdr *) pSMBr, &bytes_returned, timeout);
+
+		if (rc) {
+			cERROR(1, ("Send error in Lock = %d", rc));
+		}
+		if (pSMB)
+			buf_release(pSMB);
+
+		/* Note: On -EAGAIN error only caller can retry on handle based calls
+		since file handle passed in no longer valid */
 
-	rc = smb_init(SMB_COM_LOCKING_ANDX, 8, tcon, (void **) &pSMB,
-		      (void **) &pSMBr);
-	if (rc)
 		return rc;
-
-	if(lockType == LOCKING_ANDX_OPLOCK_RELEASE)
-		timeout = -1; /* no response expected */
-
-	pSMB->NumberOfLocks = cpu_to_le32(numLock);
-	pSMB->NumberOfUnlocks = cpu_to_le32(numUnlock);
-	pSMB->LockType = lockType;
-	pSMB->AndXCommand = 0xFF;	/* none */
-	pSMB->Fid = smb_file_id; /* netfid stays le */
-
-	pSMB->Locks[0].Pid = cpu_to_le16(current->pid);
-	pSMB->Locks[0].Length = cpu_to_le64(len);
-	pSMB->Locks[0].Offset = cpu_to_le64(offset);
-	pSMB->ByteCount = sizeof (LOCKING_ANDX_RANGE);
-	pSMB->hdr.smb_buf_length += pSMB->ByteCount;
-	pSMB->ByteCount = cpu_to_le16(pSMB->ByteCount);
-
-	rc = SendReceive(xid, tcon->ses, (struct smb_hdr *) pSMB,
-			 (struct smb_hdr *) pSMBr, &bytes_returned, timeout);
-
-	if (rc) {
-		cERROR(1, ("Send error in Lock = %d", rc));
-	}
-	if (pSMB)
-		buf_release(pSMB);
-
-	return rc;
 }
 
 int
@@ -676,8 +807,11 @@ CIFSSMBClose(const int xid, struct cifsTconInfo *tcon, int smb_file_id)
 	int bytes_returned;
 	cFYI(1, ("In CIFSSMBClose"));
 
+/* do not retry on dead session on close */
 	rc = smb_init(SMB_COM_CLOSE, 3, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
+	if(rc == -EAGAIN)
+		return 0;
 	if (rc)
 		return rc;
 
@@ -692,6 +826,10 @@ CIFSSMBClose(const int xid, struct cifsTconInfo *tcon, int smb_file_id)
 	if (pSMB)
 		buf_release(pSMB);
 
+	/* Since session is dead, file will be closed on server already */
+	if(rc == -EAGAIN)
+		rc = 0;
+
 	return rc;
 }
 
@@ -707,7 +845,7 @@ CIFSSMBRename(const int xid, struct cifsTconInfo *tcon,
 	int name_len, name_len2;
 
 	cFYI(1, ("In CIFSSMBRename"));
-
+renameRetry:
 	rc = smb_init(SMB_COM_RENAME, 1, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -754,14 +892,94 @@ CIFSSMBRename(const int xid, struct cifsTconInfo *tcon,
 	rc = SendReceive(xid, tcon->ses, (struct smb_hdr *) pSMB,
 			 (struct smb_hdr *) pSMBr, &bytes_returned, 0);
 	if (rc) {
-		cFYI(1, ("Send error in RMDir = %d", rc));
+		cFYI(1, ("Send error in rename = %d", rc));
 	}
 	if (pSMB)
 		buf_release(pSMB);
 
+	if (rc == -EAGAIN)
+		goto renameRetry;
+
 	return rc;
 }
 
+int CIFSSMBRenameOpenFile(const int xid,struct cifsTconInfo *pTcon, 
+		int netfid, char * target_name, const struct nls_table * nls_codepage) 
+{
+	struct smb_com_transaction2_sfi_req *pSMB  = NULL;
+	struct smb_com_transaction2_sfi_rsp *pSMBr = NULL;
+	struct set_file_rename * rename_info;
+	char *data_offset;
+	char dummy_string[30];
+	int rc = 0;
+	int bytes_returned = 0;
+	int len_of_str;
+
+        cFYI(1, ("Rename to File by handle"));
+        rc = smb_init(SMB_COM_TRANSACTION2, 15, pTcon, (void **) &pSMB,
+                      (void **) &pSMBr);
+        if (rc)
+                return rc;
+
+        pSMB->ParameterCount = 6;
+        pSMB->MaxSetupCount = 0;
+        pSMB->Reserved = 0;
+        pSMB->Flags = 0;
+        pSMB->Timeout = 0;
+        pSMB->Reserved2 = 0;
+        pSMB->ParameterOffset = offsetof(struct smb_com_transaction2_sfi_req,
+                                     Fid) - 4;
+        pSMB->DataOffset = pSMB->ParameterOffset + pSMB->ParameterCount;
+
+        data_offset = (char *) (&pSMB->hdr.Protocol) + pSMB->DataOffset;
+	rename_info = (struct set_file_rename *) data_offset;
+        pSMB->MaxParameterCount = cpu_to_le16(2);
+        pSMB->MaxDataCount = cpu_to_le16(1000); /* BB find max SMB PDU from sess */
+        pSMB->SetupCount = 1;
+        pSMB->Reserved3 = 0;
+        pSMB->SubCommand = cpu_to_le16(TRANS2_SET_FILE_INFORMATION);
+        pSMB->ByteCount = 3 /* pad */  + pSMB->ParameterCount;
+        pSMB->ParameterCount = cpu_to_le16(pSMB->ParameterCount);
+        pSMB->TotalParameterCount = pSMB->ParameterCount;
+        pSMB->ParameterOffset = cpu_to_le16(pSMB->ParameterOffset);
+        pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset);
+	/* construct random name ".cifs_tmp<inodenum><mid>" */
+	rename_info->overwrite = cpu_to_le32(1);
+	rename_info->root_fid  = 0;
+	/* unicode only call */
+	if(target_name == NULL) {
+		sprintf(dummy_string,"cifs%x",pSMB->hdr.Mid);
+	        len_of_str = cifs_strtoUCS((wchar_t *) rename_info->target_name, dummy_string, 24, nls_codepage);
+	} else {
+		len_of_str = cifs_strtoUCS((wchar_t *) rename_info->target_name, target_name, 530, nls_codepage);
+	}
+	rename_info->target_name_len = cpu_to_le32(2 * len_of_str);
+	pSMB->DataCount = 12 /* sizeof(struct set_file_rename) */ + (2 * len_of_str) + 2;
+	pSMB->ByteCount += pSMB->DataCount;
+	pSMB->DataCount = cpu_to_le16(pSMB->DataCount);
+	pSMB->TotalDataCount = pSMB->DataCount;
+	pSMB->Fid = netfid;
+	pSMB->InformationLevel =
+		cpu_to_le16(SMB_SET_FILE_RENAME_INFORMATION);
+	pSMB->Reserved4 = 0;
+	pSMB->hdr.smb_buf_length += pSMB->ByteCount;
+	pSMB->ByteCount = cpu_to_le16(pSMB->ByteCount);
+	rc = SendReceive(xid, pTcon->ses, (struct smb_hdr *) pSMB,
+                         (struct smb_hdr *) pSMBr, &bytes_returned, 0);
+	if (rc) {
+                cFYI(1,("Send error in Rename (by file handle) = %d", rc));
+	}
+
+	if (pSMB)
+                buf_release(pSMB);
+
+	/* Note: On -EAGAIN error only caller can retry on handle based calls
+		since file handle passed in no longer valid */
+
+	return rc;
+}
+
+
 int
 CIFSUnixCreateSymLink(const int xid, struct cifsTconInfo *tcon,
 		      const char *fromName, const char *toName,
@@ -776,7 +994,7 @@ CIFSUnixCreateSymLink(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In Symlink Unix style"));
-
+createSymLinkRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -847,6 +1065,10 @@ CIFSUnixCreateSymLink(const int xid, struct cifsTconInfo *tcon,
 
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto createSymLinkRetry;
+
 	return rc;
 }
 
@@ -864,7 +1086,7 @@ CIFSUnixCreateHardLink(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In Create Hard link Unix style"));
-
+createHardLinkRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -932,6 +1154,9 @@ CIFSUnixCreateHardLink(const int xid, struct cifsTconInfo *tcon,
 
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto createHardLinkRetry;
+
 	return rc;
 }
 
@@ -947,6 +1172,7 @@ CIFSCreateHardLink(const int xid, struct cifsTconInfo *tcon,
 	int name_len, name_len2;
 
 	cFYI(1, ("In CIFSCreateHardLink"));
+winCreateHardLinkRetry:
 
 	rc = smb_init(SMB_COM_NT_RENAME, 4, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
@@ -999,6 +1225,8 @@ CIFSCreateHardLink(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto winCreateHardLinkRetry;
 
 	return rc;
 }
@@ -1017,6 +1245,8 @@ CIFSSMBUnixQuerySymLink(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In QPathSymLinkInfo (Unix) for path %s", searchName));
+
+querySymLinkRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1073,18 +1303,31 @@ CIFSSMBUnixQuerySymLink(const int xid, struct cifsTconInfo *tcon,
 			rc = -EIO;	/* bad smb */
 		else {
 			if (pSMBr->hdr.Flags2 & SMBFLG2_UNICODE) {
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 				name_len = UniStrnlen((wchar_t *) ((char *)
 					&pSMBr->hdr.Protocol +pSMBr->DataOffset),
 					min_t(const int, buflen,pSMBr->DataCount) / 2);
+#else
+                                name_len = UniStrnlen((wchar_t *) ((char *)
+                                        &pSMBr->hdr.Protocol +pSMBr->DataOffset),
+                                        min(buflen,pSMBr->DataCount) / 2);
+#endif
 				cifs_strfromUCS_le(symlinkinfo,
 					(wchar_t *) ((char *)&pSMBr->hdr.Protocol +
 						pSMBr->DataOffset),
 					name_len, nls_codepage);
 			} else {
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 				strncpy(symlinkinfo,
 					(char *) &pSMBr->hdr.Protocol + 
 						pSMBr->DataOffset,
 					min_t(const int, buflen, pSMBr->DataCount));
+#else
+                                strncpy(symlinkinfo,
+                                        (char *) &pSMBr->hdr.Protocol +
+                                                pSMBr->DataOffset,
+                                        min(buflen, pSMBr->DataCount));
+#endif
 			}
 			symlinkinfo[buflen] = 0;
 	/* just in case so calling code does not go off the end of buffer */
@@ -1092,6 +1335,8 @@ CIFSSMBUnixQuerySymLink(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto querySymLinkRetry;
 	return rc;
 }
 
@@ -1159,9 +1404,15 @@ CIFSSMBQueryReparseLinkInfo(const int xid, struct cifsTconInfo *tcon,
 						reparse_buf->TargetNameOffset),
 						name_len, nls_codepage);
 				} else { /* ASCII names */
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 					strncpy(symlinkinfo,reparse_buf->LinkNamesBuf + 
 						reparse_buf->TargetNameOffset, 
 						min_t(const int, buflen, reparse_buf->TargetNameLen));
+#else
+                                        strncpy(symlinkinfo,reparse_buf->LinkNamesBuf +
+                                                reparse_buf->TargetNameOffset,
+                                                min(buflen, reparse_buf->TargetNameLen));
+#endif
 				}
 			} else {
 				rc = -EIO;
@@ -1174,6 +1425,10 @@ CIFSSMBQueryReparseLinkInfo(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	/* Note: On -EAGAIN error only caller can retry on handle based calls
+		since file handle passed in no longer valid */
+
 	return rc;
 }
 
@@ -1191,6 +1446,7 @@ CIFSSMBQPathInfo(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In QPathInfo path %s", searchName));
+QPathInfoRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1252,6 +1508,9 @@ CIFSSMBQPathInfo(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto QPathInfoRetry;
+
 	return rc;
 }
 
@@ -1269,6 +1528,7 @@ CIFSSMBUnixQPathInfo(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In QPathInfo (Unix) the path %s", searchName));
+UnixQPathInfoRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1331,6 +1591,9 @@ CIFSSMBUnixQPathInfo(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto UnixQPathInfoRetry;
+
 	return rc;
 }
 
@@ -1347,6 +1610,7 @@ CIFSFindSingle(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In FindUnique"));
+findUniqueRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1405,6 +1669,9 @@ CIFSFindSingle(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto findUniqueRetry;
+
 	return rc;
 }
 
@@ -1424,6 +1691,7 @@ CIFSFindFirst(const int xid, struct cifsTconInfo *tcon,
 	int name_len;
 
 	cFYI(1, ("In FindFirst"));
+findFirstRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1508,6 +1776,10 @@ CIFSFindFirst(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto findFirstRetry;
+
 	return rc;
 }
 
@@ -1525,6 +1797,7 @@ CIFSFindNext(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned;
 
 	cFYI(1, ("In FindNext"));
+
 	if(resume_file_name == NULL) {
 		return -EIO;
 	}
@@ -1608,6 +1881,10 @@ CIFSFindNext(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	/* Note: On -EAGAIN error only caller can retry on handle based calls
+		since file handle passed in no longer valid */
+
 	return rc;
 }
 
@@ -1618,10 +1895,14 @@ CIFSFindClose(const int xid, struct cifsTconInfo *tcon, const __u16 searchHandle
 	FINDCLOSE_REQ *pSMB = NULL;
 	CLOSE_RSP *pSMBr = NULL;
 	int bytes_returned;
-	cFYI(1, ("In CIFSSMBFindClose"));
 
+	cFYI(1, ("In CIFSSMBFindClose"));
 	rc = smb_init(SMB_COM_FIND_CLOSE2, 1, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
+	/* no sense returning error if session restarted
+		file handle has been closed */
+	if(rc == -EAGAIN)
+		return 0;
 	if (rc)
 		return rc;
 
@@ -1635,6 +1916,10 @@ CIFSFindClose(const int xid, struct cifsTconInfo *tcon, const __u16 searchHandle
 	if (pSMB)
 		buf_release(pSMB);
 
+	/* Since session is dead, search handle closed on server already */
+	if (rc == -EAGAIN)
+		rc = 0;
+
 	return rc;
 }
 
@@ -1660,7 +1945,7 @@ CIFSGetDFSRefer(const int xid, struct cifsSesInfo *ses,
 	cFYI(1, ("In GetDFSRefer the path %s", searchName));
 	if (ses == NULL)
 		return -ENODEV;
-
+getDFSRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, 0, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1792,6 +2077,10 @@ CIFSGetDFSRefer(const int xid, struct cifsSesInfo *ses,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto getDFSRetry;
+
 	return rc;
 }
 
@@ -1807,7 +2096,7 @@ CIFSSMBQFSInfo(const int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In QFSInfo"));
-
+QFSInfoRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1868,6 +2157,10 @@ CIFSSMBQFSInfo(const int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto QFSInfoRetry;
+
 	return rc;
 }
 
@@ -1883,6 +2176,7 @@ CIFSSMBQFSAttributeInfo(int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In QFSAttributeInfo"));
+QFSAttributeRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1930,6 +2224,10 @@ CIFSSMBQFSAttributeInfo(int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto QFSAttributeRetry;
+
 	return rc;
 }
 
@@ -1945,7 +2243,7 @@ CIFSSMBQFSDeviceInfo(int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In QFSDeviceInfo"));
-
+QFSDeviceRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -1978,7 +2276,7 @@ CIFSSMBQFSDeviceInfo(int xid, struct cifsTconInfo *tcon,
 	rc = SendReceive(xid, tcon->ses, (struct smb_hdr *) pSMB,
 			 (struct smb_hdr *) pSMBr, &bytes_returned, 0);
 	if (rc) {
-		cERROR(1, ("Send error in QFSDeviceInfo = %d", rc));
+		cFYI(1, ("Send error in QFSDeviceInfo = %d", rc));
 	} else {		/* decode response */
 		pSMBr->DataOffset = le16_to_cpu(pSMBr->DataOffset);
 		if ((pSMBr->ByteCount < sizeof (FILE_SYSTEM_DEVICE_INFO))
@@ -1995,6 +2293,11 @@ CIFSSMBQFSDeviceInfo(int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto QFSDeviceRetry;
+
+
 	return rc;
 }
 
@@ -2010,6 +2313,7 @@ CIFSSMBQFSUnixInfo(int xid, struct cifsTconInfo *tcon,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In QFSUnixInfo"));
+QFSUnixRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -2057,6 +2361,11 @@ CIFSSMBQFSUnixInfo(int xid, struct cifsTconInfo *tcon,
 	}
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto QFSUnixRetry;
+
+
 	return rc;
 }
 
@@ -2078,7 +2387,7 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 	int bytes_returned = 0;
 
 	cFYI(1, ("In SetEOF"));
-
+SetEOFRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -2149,6 +2458,10 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto SetEOFRetry;
+
 	return rc;
 }
 
@@ -2165,7 +2478,6 @@ CIFSSMBSetFileSize(const int xid, struct cifsTconInfo *tcon, __u64 size,
 	__u32 tmp;
 
 	cFYI(1, ("SetFileSize (via SetFileInfo)"));
-
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -2201,10 +2513,10 @@ CIFSSMBSetFileSize(const int xid, struct cifsTconInfo *tcon, __u64 size,
 	pSMB->TotalDataCount = pSMB->DataCount;
 	pSMB->TotalParameterCount = pSMB->ParameterCount;
 	pSMB->ParameterOffset = cpu_to_le16(pSMB->ParameterOffset);
-	pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset);
 	parm_data =
 		(struct file_end_of_file_info *) (((char *) &pSMB->hdr.Protocol) +
 			pSMB->DataOffset);
+	pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset); /* now safe to change to le */
 	parm_data->FileSize = size;
 	pSMB->Fid = fid;
 	if(SetAllocation) {
@@ -2235,6 +2547,10 @@ CIFSSMBSetFileSize(const int xid, struct cifsTconInfo *tcon, __u64 size,
 
 	if (pSMB)
 		buf_release(pSMB);
+
+ 	/* Note: On -EAGAIN error only caller can retry on handle based calls 
+		since file handle passed in no longer valid */
+
 	return rc;
 }
 
@@ -2251,6 +2567,7 @@ CIFSSMBSetTimes(int xid, struct cifsTconInfo *tcon, char *fileName,
 
 	cFYI(1, ("In SetTimes"));
 
+SetTimesRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -2309,6 +2626,10 @@ CIFSSMBSetTimes(int xid, struct cifsTconInfo *tcon, char *fileName,
 
 	if (pSMB)
 		buf_release(pSMB);
+
+	if (rc == -EAGAIN)
+		goto SetTimesRetry;
+
 	return rc;
 }
 
@@ -2325,7 +2646,7 @@ CIFSSMBUnixSetPerms(const int xid, struct cifsTconInfo *tcon,
 	FILE_UNIX_BASIC_INFO *data_offset;
 
 	cFYI(1, ("In SetUID/GID/Mode"));
-
+setPermsRetry:
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
@@ -2387,5 +2708,7 @@ CIFSSMBUnixSetPerms(const int xid, struct cifsTconInfo *tcon,
 
 	if (pSMB)
 		buf_release(pSMB);
+	if (rc == -EAGAIN)
+		goto setPermsRetry;
 	return rc;
 }
