@@ -3,7 +3,7 @@
  *
  *   vfs operations that deal with files
  * 
- *   Copyright (c) International Business Machines  Corp., 2002
+ *   Copyright (C) International Business Machines  Corp., 2002,2003
  *   Author(s): Steve French (sfrench@us.ibm.com)
  *
  *   This library is free software; you can redistribute it and/or modify
@@ -26,6 +26,7 @@
 #include <linux/version.h>
 #include <linux/pagemap.h>
 #include <linux/smp_lock.h>
+#include <linux/list.h>
 #include <asm/div64.h>
 #include <linux/mm.h>
 #include "cifsfs.h"
@@ -47,7 +48,7 @@ cifs_open(struct inode *inode, struct file *file)
 	struct cifsInodeInfo *pCifsInode;
 	char *full_path = NULL;
 	int desiredAccess = 0x20197;
-	int disposition = FILE_OPEN;
+	int disposition;
 	__u16 netfid;
 	FILE_ALL_INFO * buf = NULL;
 
@@ -66,26 +67,38 @@ cifs_open(struct inode *inode, struct file *file)
 	else if ((file->f_flags & O_ACCMODE) == O_RDWR)
 		desiredAccess = GENERIC_ALL;
 
-/* BB check other flags carefully to find equivalent NTCreateX flags */
+/*********************************************************************
+ *  open flag mapping table:
+ *  
+ *	POSIX Flag            CIFS Disposition
+ *	----------            ---------------- 
+ *	O_CREAT               FILE_OPEN_IF
+ *	O_CREAT | O_EXCL      FILE_CREATE
+ *	O_CREAT | O_TRUNC     FILE_OVERWRITE_IF
+ *	O_TRUNC               FILE_OVERWRITE
+ *	none of the above     FILE_OPEN
+ *
+ *	Note that there is not a direct match between disposition
+ *	FILE_SUPERSEDE (ie create whether or not file exists although 
+ *	O_CREAT | O_TRUNC is similar but truncates the existing
+ *	file rather than creating a new file as FILE_SUPERSEDE does
+ *	(which uses the attributes / metadata passed in on open call)
+ *?
+ *?  O_SYNC is a reasonable match to CIFS writethrough flag  
+ *?  and the read write flags match reasonably.  O_LARGEFILE
+ *?  is irrelevant because largefile support is always used
+ *?  by this client. Flags O_APPEND, O_DIRECT, O_DIRECTORY,
+ *	 O_FASYNC, O_NOFOLLOW, O_NONBLOCK need further investigation
+ *********************************************************************/
 
-/*
-#define O_CREAT		   0100	
-#define O_EXCL		   0200	
-#define O_NOCTTY	   0400	
-#define O_TRUNC		  01000	
-#define O_APPEND	  02000
-#define O_NONBLOCK	  04000
-#define O_NDELAY	O_NONBLOCK
-#define O_SYNC		 010000
-#define FASYNC		 020000	
-#define O_DIRECT	 040000	
-#define O_LARGEFILE	0100000
-#define O_DIRECTORY	0200000	
-#define O_NOFOLLOW	0400000
-#define O_ATOMICLOOKUP	01000000 */
-
-	if (file->f_flags & O_CREAT)
-		disposition = FILE_OVERWRITE;
+	if((file->f_flags & (O_CREAT | O_EXCL)) == (O_CREAT | O_EXCL))
+		disposition = FILE_CREATE;
+	else if((file->f_flags & (O_CREAT | O_TRUNC)) == (O_CREAT | O_TRUNC))
+		disposition = FILE_OVERWRITE_IF;
+	else if((file->f_flags & O_CREAT) == O_CREAT)
+		disposition = FILE_OPEN_IF;
+	else
+		disposition = FILE_OPEN;
 
 	if (oplockEnabled)
 		oplock = REQ_OPLOCK;
@@ -97,7 +110,14 @@ cifs_open(struct inode *inode, struct file *file)
 	/* Also refresh inode by passing in file_info buf returned by SMBOpen 
 	   and calling get_inode_info with returned buf (at least 
 	   helps non-Unix server case */
-        buf = kmalloc(sizeof(FILE_ALL_INFO),GFP_KERNEL);
+	buf = kmalloc(sizeof(FILE_ALL_INFO),GFP_KERNEL);
+	if(buf==0) {
+		if (full_path)
+			kfree(full_path);
+		FreeXid(xid);
+		return -ENOMEM;
+	}
+
 	rc = CIFSSMBOpen(xid, pTcon, full_path, disposition, desiredAccess,
 			CREATE_NOT_DIR, &netfid, &oplock, buf, cifs_sb->local_nls);
 	if (rc) {
@@ -112,13 +132,20 @@ cifs_open(struct inode *inode, struct file *file)
 			pCifsFile = (struct cifsFileInfo *) file->private_data;
 			pCifsFile->netfid = netfid;
 			pCifsFile->pid = current->pid;
+			init_MUTEX(&pCifsFile->fh_sem);
 			pCifsFile->pfile = file; /* needed for writepage */
 			pCifsFile->pInode = inode;
-			spin_lock(&files_lock);
+			pCifsFile->invalidHandle = FALSE;
+			pCifsFile->closePend     = FALSE;
 			write_lock(&GlobalSMBSeslock);
+			spin_lock(&files_lock);
 			list_add(&pCifsFile->tlist,&pTcon->openFileList);
 			pCifsInode = CIFS_I(file->f_dentry->d_inode);
 			if(pCifsInode) {
+				list_add(&pCifsFile->flist,&pCifsInode->openFileList);
+				write_unlock(&files_lock);
+				write_unlock(&GlobalSMBSeslock);
+
 		                if (pTcon->ses->capabilities & CAP_UNIX)
 					rc = cifs_get_inode_info_unix(&file->f_dentry->d_inode,
 						full_path, inode->i_sb);
@@ -126,16 +153,17 @@ cifs_open(struct inode *inode, struct file *file)
 					rc = cifs_get_inode_info(&file->f_dentry->d_inode,
 						full_path, buf, inode->i_sb);
 
-				list_add(&pCifsFile->flist,&pCifsInode->openFileList);
 				if(oplock == OPLOCK_EXCLUSIVE) {
 					pCifsInode->clientCanCacheAll = TRUE;
 					pCifsInode->clientCanCacheRead = TRUE;
 					cFYI(1,("Exclusive Oplock granted on inode %p",file->f_dentry->d_inode));
 				} else if(oplock == OPLOCK_READ)
 					pCifsInode->clientCanCacheRead = TRUE;
+			} else {
+				write_unlock(&files_lock);
+				write_unlock(&GlobalSMBSeslock);
 			}
-			write_unlock(&GlobalSMBSeslock);
-			spin_unlock(&files_lock);
+
 			if(file->f_flags & O_CREAT) {           
 				/* time to set mode which we can not set earlier due
 				 to problems creating new read-only files */
@@ -172,44 +200,106 @@ int relock_files(struct cifsFileInfo * cifsFile)
 	return rc;
 }
 
-/* Try to reopen files that were closed when session to server was lost */
-int reopen_files(struct cifsTconInfo * pTcon, struct nls_table * nlsinfo)
+static int cifs_reopen_file(struct inode *inode, struct file *file)
 {
-	int rc = 0;
-	struct cifsFileInfo *open_file = NULL;
-	struct file * file = NULL;
-	struct list_head *tmp;
-	struct list_head *tmp1;
+        int rc = -EACCES;
+        int xid, oplock;
+        struct cifs_sb_info *cifs_sb;
+        struct cifsTconInfo *pTcon;
+        struct cifsFileInfo *pCifsFile;
+        struct cifsInodeInfo *pCifsInode;
+        char *full_path = NULL;
+        int desiredAccess = 0x20197;
+        int disposition = FILE_OPEN;
+        __u16 netfid;
+        FILE_ALL_INFO * buf = NULL;
 
-/* list all files open on tree connection */
-	read_lock(&GlobalSMBSeslock);
-	list_for_each_safe(tmp, tmp1, &pTcon->openFileList) {            
-		open_file = list_entry(tmp,struct cifsFileInfo, tlist);
-		if(open_file) {
-			if(open_file->search_resume_name) {
-				kfree(open_file->search_resume_name);
-			}
-			file = open_file->pfile;
-			list_del(&open_file->flist);
-			list_del(&open_file->tlist);
-			kfree(open_file);
-			if(file) {                
-				file->private_data = NULL;
-				read_unlock(&GlobalSMBSeslock);
-				rc = cifs_open(file->f_dentry->d_inode,file);
-				read_lock(&GlobalSMBSeslock);
-				if(rc) {
-					cFYI(1,("reconnecting file %s failed with %d",
-						file->f_dentry->d_name.name,rc));
-				} else {
-					cFYI(1,("reconnection of %s succeeded",
-						file->f_dentry->d_name.name));
-				}
-			} 
-		}
-	}
-	read_unlock(&GlobalSMBSeslock);
-	return rc;
+        if(inode == NULL)
+                return -EBADF;
+        if (file->private_data) {
+                pCifsFile = (struct cifsFileInfo *) file->private_data;
+        } else
+                return -EBADF;
+
+        xid = GetXid();
+        down(&pCifsFile->fh_sem);
+        if(pCifsFile->invalidHandle == FALSE) {
+                up(&pCifsFile->fh_sem);
+                FreeXid(xid);
+                return 0;
+        }
+
+
+        cifs_sb = CIFS_SB(inode->i_sb);
+        pTcon = cifs_sb->tcon;
+
+        full_path = build_path_from_dentry(file->f_dentry);
+
+        cFYI(1, (" inode = 0x%p file flags are 0x%x for %s", inode, file->f_flags,full_path));
+        if ((file->f_flags & O_ACCMODE) == O_RDONLY)
+                desiredAccess = GENERIC_READ;
+        else if ((file->f_flags & O_ACCMODE) == O_WRONLY)
+                desiredAccess = GENERIC_WRITE;
+        else if ((file->f_flags & O_ACCMODE) == O_RDWR)
+                desiredAccess = GENERIC_ALL;
+
+        if (oplockEnabled)
+                oplock = REQ_OPLOCK;
+        else
+                oplock = FALSE;
+
+        /* BB pass O_SYNC flag through on file attributes .. BB */
+
+        /* Also refresh inode by passing in file_info buf returned by SMBOpen
+           and calling get_inode_info with returned buf (at least
+           helps non-Unix server case */
+        buf = kmalloc(sizeof(FILE_ALL_INFO),GFP_KERNEL);
+        if(buf==0) {
+                up(&pCifsFile->fh_sem);
+                if (full_path)
+                        kfree(full_path);
+                FreeXid(xid);
+                return -ENOMEM;
+        }
+        rc = CIFSSMBOpen(xid, pTcon, full_path, disposition, desiredAccess,
+                                CREATE_NOT_DIR, &netfid, &oplock, buf, cifs_sb->local_nls);
+        if (rc) {
+                up(&pCifsFile->fh_sem);
+                cFYI(1, ("cifs_open returned 0x%x ", rc));
+                cFYI(1, ("oplock: %d ", oplock));
+        } else {
+                pCifsFile->netfid = netfid;
+                pCifsFile->invalidHandle = FALSE;
+                up(&pCifsFile->fh_sem);
+                pCifsInode = CIFS_I(inode);
+                if(pCifsInode) {
+                        if (pTcon->ses->capabilities & CAP_UNIX)
+                                rc = cifs_get_inode_info_unix(&inode,
+                                                full_path, inode->i_sb);
+                        else
+                                rc = cifs_get_inode_info(&inode,
+                                                full_path, buf, inode->i_sb);
+
+                        if(oplock == OPLOCK_EXCLUSIVE) {
+                                pCifsInode->clientCanCacheAll =  TRUE;
+                                pCifsInode->clientCanCacheRead = TRUE;
+                                cFYI(1,("Exclusive Oplock granted on inode %p",file->f_dentry->d_inode));
+                        } else if(oplock == OPLOCK_READ) {
+                                pCifsInode->clientCanCacheRead = TRUE;
+                                pCifsInode->clientCanCacheAll =  FALSE;
+                        } else {
+                                pCifsInode->clientCanCacheRead = FALSE;
+                                pCifsInode->clientCanCacheAll =  FALSE;
+                        }
+                }
+        }
+
+        if (buf)
+                kfree(buf);
+        if (full_path)
+                kfree(full_path);
+        FreeXid(xid);
+        return rc;
 }
 
 int
@@ -227,11 +317,20 @@ cifs_close(struct inode *inode, struct file *file)
 	cifs_sb = CIFS_SB(inode->i_sb);
 	pTcon = cifs_sb->tcon;
 	if (pSMBFile) {
+		pSMBFile->closePend	= TRUE;
 		spin_lock(&files_lock);
+		if(pTcon) {
+			/* no sense reconnecting to close a file that is
+				already closed */
+			if (pTcon->tidStatus != CifsNeedReconnect) {
+				spin_unlock(&files_lock);
+				rc = CIFSSMBClose(xid,pTcon,pSMBFile->netfid);
+				spin_lock(&files_lock);
+			}
+		}
 		list_del(&pSMBFile->flist);
 		list_del(&pSMBFile->tlist);
 		spin_unlock(&files_lock);
-		rc = CIFSSMBClose(xid, pTcon, pSMBFile->netfid);
 		if(pSMBFile->search_resume_name)
 			kfree(pSMBFile->search_resume_name);
 		kfree(file->private_data);
@@ -245,6 +344,7 @@ cifs_close(struct inode *inode, struct file *file)
 		info on this inode, much less write behind and read ahead */
 		CIFS_I(inode)->clientCanCacheRead = FALSE;
 		CIFS_I(inode)->clientCanCacheAll  = FALSE;
+		invalidate_inode_pages(inode);
 	}
 	if((rc ==0) && CIFS_I(inode)->write_behind_rc)
 		rc = CIFS_I(inode)->write_behind_rc;
@@ -281,6 +381,7 @@ cifs_lock(struct file *file, int cmd, struct file_lock *pfLock)
 	__u32 numLock = 0;
 	__u32 numUnlock = 0;
 	__u64 length;
+	int wait_flag = FALSE;
 	struct cifs_sb_info *cifs_sb;
 	struct cifsTconInfo *pTcon;
 	length = 1 + pfLock->fl_end - pfLock->fl_start;
@@ -298,14 +399,17 @@ cifs_lock(struct file *file, int cmd, struct file_lock *pfLock)
 		cFYI(1, ("Posix "));
 	if (pfLock->fl_flags & FL_FLOCK)
 		cFYI(1, ("Flock "));
-/*	if (pfLock->fl_flags & FL_SLEEP)
-		cFYI(1, ("Blocking lock ")); */
+/*	if (pfLock->fl_flags & FL_SLEEP) {
+		cFYI(1, ("Blocking lock "));
+		wait_flag = TRUE;
+	} */                          
+
 	if (pfLock->fl_flags & FL_ACCESS)
-		cFYI(1, ("Process suspended by mandatory locking "));
+		cFYI(1, ("Process suspended by mandatory locking - not implemented yet "));
 	if (pfLock->fl_flags & FL_LEASE)
-		cFYI(1, ("Lease on file "));
-	if (pfLock->fl_flags & 0xFFD0)
-		cFYI(1, ("Unknown lock flags "));
+		cFYI(1, ("Lease on file - not implemented yet"));
+	if (pfLock->fl_flags & (~(FL_POSIX | FL_FLOCK | FL_ACCESS | FL_LEASE)))
+		cFYI(1, ("Unknown lock flags 0x%x",pfLock->fl_flags));
 
 	if (pfLock->fl_type == F_WRLCK) {
 		cFYI(1, ("F_WRLCK "));
@@ -353,7 +457,7 @@ cifs_lock(struct file *file, int cmd, struct file_lock *pfLock)
 			pfLock->fl_type = F_UNLCK;
 			if (rc != 0)
 				cERROR(1,
-				       ("Error unlocking previously locked range %d during test of lock ",
+					("Error unlocking previously locked range %d during test of lock ",
 					rc));
 			rc = 0;
 
@@ -370,7 +474,7 @@ cifs_lock(struct file *file, int cmd, struct file_lock *pfLock)
 			 ((struct cifsFileInfo *) file->private_data)->
 			 netfid, length,
 			 pfLock->fl_start, numUnlock, numLock, lockType,
-			 0 /* wait flag */ );
+			 wait_flag);
 	FreeXid(xid);
 	return rc;
 }
@@ -385,6 +489,7 @@ cifs_write(struct file * file, const char *write_data,
 	struct cifs_sb_info *cifs_sb;
 	struct cifsTconInfo *pTcon;
 	int xid, long_op;
+	struct cifsFileInfo * open_file;
 
 	xid = GetXid();
 
@@ -399,22 +504,31 @@ cifs_write(struct file * file, const char *write_data,
 		FreeXid(xid);
 		return -EBADF;
 	}
+	open_file = (struct cifsFileInfo *) file->private_data;
+
 
 	if (*poffset > file->f_dentry->d_inode->i_size)
-		long_op = 2;	/* writes past end of file can take a long time */
+		long_op = 2;  /* writes past end of file can take a long time */
 	else
 		long_op = 1;
 
 	for (total_written = 0; write_size > total_written;
 	     total_written += bytes_written) {
-		rc = CIFSSMBWrite(xid, pTcon,
-				  ((struct cifsFileInfo *) file->
-				   private_data)->netfid,
+		rc = -EAGAIN;
+		while(rc == -EAGAIN) {
+			if ((open_file->invalidHandle) && (!open_file->closePend)) {
+				rc = cifs_reopen_file(file->f_dentry->d_inode,file);
+				if(rc != 0)
+					break;
+			}
+
+			rc = CIFSSMBWrite(xid, pTcon,
+				   open_file->netfid,
 				  write_size - total_written, *poffset,
 				  &bytes_written,
 				  write_data + total_written, long_op);
+		}
 		if (rc || (bytes_written == 0)) {
-			FreeXid(xid);
 			if (total_written)
 				break;
 			else {
@@ -425,7 +539,8 @@ cifs_write(struct file * file, const char *write_data,
 			*poffset += bytes_written;
 		long_op = FALSE; /* subsequent writes fast - 15 seconds is plenty */
 	}
-	file->f_dentry->d_inode->i_ctime = file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
+	file->f_dentry->d_inode->i_ctime = file->f_dentry->d_inode->i_mtime =
+		CURRENT_TIME;
 	if (bytes_written > 0) {
 		if (*poffset > file->f_dentry->d_inode->i_size)
 			file->f_dentry->d_inode->i_size = *poffset;
@@ -450,24 +565,19 @@ cifs_partialpagewrite(struct page *page,unsigned from, unsigned to)
 	struct cifsFileInfo *open_file = NULL;
 	struct list_head *tmp;
 	struct list_head *tmp1;
-	int xid;
-
-	xid = GetXid();
 
 	cifs_sb = CIFS_SB(inode->i_sb);
 	pTcon = cifs_sb->tcon;
 
 	/* figure out which file struct to use 
 	if (file->private_data == NULL) {
-	   FreeXid(xid);
+	   kunmap(page);
 	   return -EBADF;
 	}     
 	 */
 	if (!mapping) {
-		FreeXid(xid);
 		return -EFAULT;
 	} else if(!mapping->host) {
-		FreeXid(xid);
 		return -EFAULT;
 	}
 
@@ -475,12 +585,12 @@ cifs_partialpagewrite(struct page *page,unsigned from, unsigned to)
 	write_data += from;
 
 	if((to > PAGE_CACHE_SIZE) || (from > to) || (offset > mapping->host->i_size)) {
-		FreeXid(xid);
+		kunmap(page);
 		return -EIO;
 	}
 
 	/* check to make sure that we are not extending the file */
-    if(mapping->host->i_size - offset < (loff_t)to)
+	if(mapping->host->i_size - offset < (loff_t)to)
 		to = (unsigned)(mapping->host->i_size - offset); 
 		
 
@@ -503,6 +613,13 @@ cifs_partialpagewrite(struct page *page,unsigned from, unsigned to)
 			} else if(bytes_written < 0) {
 				rc = bytes_written;
 			}
+			break;  /* now that we found a valid file handle
+                                and tried to write to it we are done, no
+                                sense continuing to loop looking for another */
+		}
+		if(tmp->next == NULL) {
+			cFYI(1,("File instance %p removed",tmp));
+			break;
 		}
 	}
 	read_unlock(&GlobalSMBSeslock);
@@ -511,7 +628,7 @@ cifs_partialpagewrite(struct page *page,unsigned from, unsigned to)
 		rc = -EIO;
 	}
 
-	FreeXid(xid);
+	kunmap(page);
 	return rc;
 }
 
@@ -541,7 +658,11 @@ cifs_writepage(struct page* page)
 	
 	rc = cifs_partialpagewrite(page,0,PAGE_CACHE_SIZE);
 	SetPageUptodate(page); /* BB add check for error and Clearuptodate? */
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,20))
 	unlock_page(page);
+#else
+	UnlockPage(page);
+#endif
 	page_cache_release(page);	
 	FreeXid(xid);
 	return rc;
@@ -565,13 +686,28 @@ cifs_commit_write(struct file *file, struct page *page, unsigned offset,
 		if (file->private_data == NULL) {
 			rc = -EBADF;
 		} else {
-			cifs_sb = CIFS_SB(inode->i_sb);
 			open_file = (struct cifsFileInfo *)file->private_data;
-			rc = CIFSSMBSetFileSize(xid, cifs_sb->tcon, position,
-				open_file->netfid,open_file->pid,FALSE);
+			cifs_sb = CIFS_SB(inode->i_sb);
+			rc = -EAGAIN;
+			while(rc == -EAGAIN) {
+				if((open_file->invalidHandle) && 
+				  (!open_file->closePend)) {
+					rc = cifs_reopen_file(file->f_dentry->d_inode,file);
+					if(rc != 0)
+						break;
+				}
+                if(!open_file->closePend)
+				    rc = CIFSSMBSetFileSize(xid, cifs_sb->tcon, 
+						position, open_file->netfid,
+						open_file->pid,FALSE);
+                else {
+                    rc = -EBADF;
+                    break;
+                }
+			}
 			cFYI(1,(" SetEOF (commit write) rc = %d",rc));
 		}
-    }
+	}
 	set_page_dirty(page);
 
 	FreeXid(xid);
@@ -674,6 +810,7 @@ cifs_read(struct file * file, char *read_data, size_t read_size,
 	struct cifsTconInfo *pTcon;
 	int xid;
 	char * current_offset;
+	struct cifsFileInfo * open_file;
 
 	xid = GetXid();
 	cifs_sb = CIFS_SB(file->f_dentry->d_sb);
@@ -683,15 +820,28 @@ cifs_read(struct file * file, char *read_data, size_t read_size,
 		FreeXid(xid);
 		return -EBADF;
 	}
+	open_file = (struct cifsFileInfo *)file->private_data;
 
 	for (total_read = 0,current_offset=read_data; read_size > total_read;
-				total_read += bytes_read,current_offset+=bytes_read) {
+		total_read += bytes_read,current_offset+=bytes_read) {
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 		current_read_size = min_t(const int,read_size - total_read,cifs_sb->rsize);
-		rc = CIFSSMBRead(xid, pTcon,
-				 ((struct cifsFileInfo *) file->
-				  private_data)->netfid,
-				 current_read_size, *poffset,
-				 &bytes_read, &current_offset);
+#else
+                current_read_size = min(read_size - total_read,cifs_sb->rsize);
+#endif
+		rc = -EAGAIN;
+		while(rc == -EAGAIN) {
+			if ((open_file->invalidHandle) && (!open_file->closePend)) {
+				rc = cifs_reopen_file(file->f_dentry->d_inode,file);
+				if(rc != 0)
+				break;
+			}
+
+			rc = CIFSSMBRead(xid, pTcon,
+				open_file->netfid,
+				current_read_size, *poffset,
+				&bytes_read, &current_offset);
+		}
 		if (rc || (bytes_read == 0)) {
 			if (total_read) {
 				break;
@@ -699,8 +849,9 @@ cifs_read(struct file * file, char *read_data, size_t read_size,
 				FreeXid(xid);
 				return rc;
 			}
-		} else
+		} else {
 			*poffset += bytes_read;
+		}
 	}
 
 	FreeXid(xid);
@@ -764,7 +915,11 @@ static void cifs_copy_cache_pages(struct address_space *mapping,
 		flush_dcache_page(page);
 		SetPageUptodate(page);
 		kunmap_atomic(page,KM_USER0);
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,20))
 		unlock_page(page);
+#else
+		UnlockPage(page);
+#endif
 		page_cache_release(page);
 		data += PAGE_CACHE_SIZE;
 	}
@@ -786,13 +941,14 @@ cifs_readpages(struct file *file, struct address_space *mapping,
 	char * smb_read_data = 0;
 	struct smb_com_read_rsp * pSMBr;
 	struct pagevec lru_pvec;
+	struct cifsFileInfo * open_file;
 
 	xid = GetXid();
 	if (file->private_data == NULL) {
 		FreeXid(xid);
 		return -EBADF;
 	}
-
+	open_file = (struct cifsFileInfo *)file->private_data;
 	cifs_sb = CIFS_SB(file->f_dentry->d_sb);
 	pTcon = cifs_sb->tcon;
 
@@ -801,6 +957,7 @@ cifs_readpages(struct file *file, struct address_space *mapping,
 	for(i = 0;i<num_pages;) {
 		spin_lock(&mapping->page_lock);
 		if(list_empty(page_list)) {
+			spin_unlock(&mapping->page_lock);
 			break;
 		}
 		page = list_entry(page_list->prev, struct page, list);
@@ -814,12 +971,24 @@ cifs_readpages(struct file *file, struct address_space *mapping,
 		
 		read_size = (num_pages - i) * PAGE_CACHE_SIZE;
 		/* Read size needs to be in multiples of one page */
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 		read_size = min_t(const unsigned int,read_size,cifs_sb->rsize & PAGE_CACHE_MASK);
-		rc = CIFSSMBRead(xid, pTcon,
-			((struct cifsFileInfo *) file->
-			 private_data)->netfid,
-			 read_size, offset,
-			 &bytes_read, &smb_read_data);
+#else
+                read_size = min(read_size,cifs_sb->rsize & PAGE_CACHE_MASK);
+#endif
+		rc = -EAGAIN;
+		while(rc == -EAGAIN) {
+			if ((open_file->invalidHandle) && (!open_file->closePend)) {
+				rc = cifs_reopen_file(file->f_dentry->d_inode,file);
+				if(rc != 0)
+					break;
+			}
+
+			rc = CIFSSMBRead(xid, pTcon,
+				open_file->netfid,
+				read_size, offset,
+				&bytes_read, &smb_read_data);
+		}
 
 		if ((rc < 0) || (smb_read_data == NULL)) {
 			cFYI(1,("Read error in readpages: %d",rc));
@@ -906,7 +1075,11 @@ cifs_readpage(struct file *file, struct page *page)
 
 io_error:
 	kunmap(page);
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,20))
 	unlock_page(page);
+#else
+	UnlockPage(page);
+#endif
 
 	page_cache_release(page);
 	FreeXid(xid);
@@ -966,8 +1139,13 @@ fill_in_inode(struct inode *tmp_inode,
 
 	/* can not fill in nlink here as in qpathinfo version and Unx search */
 	tmp_inode->i_size = pfindData->EndOfFile;
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15)) 
 	tmp_inode->i_blocks =
 		(tmp_inode->i_blksize - 1 + pfindData->AllocationSize) >> tmp_inode->i_blkbits;
+#else
+	tmp_inode->i_blocks =
+		(tmp_inode->i_blksize - 1 + pfindData->AllocationSize) >> 14;
+#endif
 	if (pfindData->AllocationSize < pfindData->EndOfFile)
 		cFYI(1, ("Possible sparse file: allocation size less than end of file "));
 	cFYI(1,
@@ -1040,8 +1218,13 @@ unix_fill_in_inode(struct inode *tmp_inode,
 	pfindData->NumOfBytes = le64_to_cpu(pfindData->NumOfBytes);
 	pfindData->EndOfFile = le64_to_cpu(pfindData->EndOfFile);
 	tmp_inode->i_size = pfindData->EndOfFile;
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 	tmp_inode->i_blocks =
                 (tmp_inode->i_blksize - 1 + pfindData->NumOfBytes) >> tmp_inode->i_blkbits;
+#else
+	tmp_inode->i_blocks =
+                (tmp_inode->i_blksize - 1 + pfindData->NumOfBytes) >> 14;
+#endif
 
 	if (S_ISREG(tmp_inode->i_mode)) {
 		cFYI(1, (" File inode "));
@@ -1063,7 +1246,7 @@ unix_fill_in_inode(struct inode *tmp_inode,
 	}
 }
 
-void
+static void
 construct_dentry(struct qstr *qstring, struct file *file,
 		 struct inode **ptmp_inode, struct dentry **pnew_dentry)
 {
@@ -1092,7 +1275,9 @@ construct_dentry(struct qstr *qstring, struct file *file,
                         can check the oplock flag on the open
                         response */
                         (*ptmp_inode)->i_blksize = CIFS_MAX_MSGSIZE;
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
                         (*ptmp_inode)->i_blkbits = 14;  /* 2**14 = CIFS_MAX_MSGSIZE */
+#endif
                         pCifsI->clientCanCacheRead = FALSE;
                         pCifsI->clientCanCacheAll = FALSE;
                         pCifsI->time = 0;
@@ -1112,7 +1297,9 @@ construct_dentry(struct qstr *qstring, struct file *file,
 			can check the oplock flag on the open
 			response */
 	                (*ptmp_inode)->i_blksize = CIFS_MAX_MSGSIZE;
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 			(*ptmp_inode)->i_blkbits = 14;  /* 2**14 = CIFS_MAX_MSGSIZE */
+#endif
 			pCifsI->clientCanCacheRead = FALSE;
 			pCifsI->clientCanCacheAll = FALSE;
 			pCifsI->time = 0;
@@ -1129,19 +1316,40 @@ construct_dentry(struct qstr *qstring, struct file *file,
 	}
 
 	tmp_dentry->d_time = jiffies;
-	(*ptmp_inode)->i_blksize =
-	    (pTcon->ses->server->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;
 	cFYI(1, ("i_blksize = %ld", (*ptmp_inode)->i_blksize));
 	*pnew_dentry = tmp_dentry;
 }
 
-void
+static void reset_resume_key(struct file * dir_file, 
+						unsigned char * filename, 
+						unsigned int len) {
+	struct cifsFileInfo *cifsFile;
+
+	cifsFile = (struct cifsFileInfo *)dir_file->private_data;
+	if(cifsFile == NULL)
+		return;
+	if(cifsFile->search_resume_name) {
+		kfree(cifsFile->search_resume_name);
+	}
+
+	cifsFile->resume_name_length = len;
+	cifsFile->search_resume_name = 
+			kmalloc(cifsFile->resume_name_length, GFP_KERNEL);
+	memcpy(cifsFile->search_resume_name, filename, 
+		   cifsFile->resume_name_length);
+	cFYI(1,("Reset resume key to: %s",filename));
+	return;
+}
+
+
+
+static int
 cifs_filldir(struct qstr *pqstring, FILE_DIRECTORY_INFO * pfindData,
 	     struct file *file, filldir_t filldir, void *direntry)
 {
 	struct inode *tmp_inode;
 	struct dentry *tmp_dentry;
-	int object_type;
+	int object_type,rc;
 
 	pqstring->name = pfindData->FileName;
 	pqstring->len = pfindData->FileNameLength;
@@ -1149,19 +1357,26 @@ cifs_filldir(struct qstr *pqstring, FILE_DIRECTORY_INFO * pfindData,
 	construct_dentry(pqstring, file, &tmp_inode, &tmp_dentry);
 
 	fill_in_inode(tmp_inode, pfindData, &object_type);
-	filldir(direntry, pfindData->FileName, pqstring->len, file->f_pos,
+	rc = filldir(direntry, pfindData->FileName, pqstring->len, file->f_pos,
 		tmp_inode->i_ino, object_type);
+	if(rc) {
+		/* due to readdir error we need to recalculate resume 
+		key so next readdir will restart on right entry */
+		reset_resume_key(file, pfindData->FileName, pqstring->len);
+		cFYI(1,("Error %d on filldir of %s",rc ,pfindData->FileName));
+	}
 	dput(tmp_dentry);
+	return rc;
 }
 
-void
+static int
 cifs_filldir_unix(struct qstr *pqstring,
 		  FILE_UNIX_INFO * pUnixFindData, struct file *file,
 		  filldir_t filldir, void *direntry)
 {
 	struct inode *tmp_inode;
 	struct dentry *tmp_dentry;
-	int object_type;
+	int object_type, rc;
 
 	pqstring->name = pUnixFindData->FileName;
 	pqstring->len = strnlen(pUnixFindData->FileName, MAX_PATHCONF);
@@ -1169,9 +1384,16 @@ cifs_filldir_unix(struct qstr *pqstring,
 	construct_dentry(pqstring, file, &tmp_inode, &tmp_dentry);
 
 	unix_fill_in_inode(tmp_inode, pUnixFindData, &object_type);
-	filldir(direntry, pUnixFindData->FileName, pqstring->len,
+	rc = filldir(direntry, pUnixFindData->FileName, pqstring->len,
 		file->f_pos, tmp_inode->i_ino, object_type);
+	if(rc) {
+		/* due to readdir error we need to recalculate resume 
+			key so next readdir will restart on right entry */
+		reset_resume_key(file, pUnixFindData->FileName,pqstring->len);
+		cFYI(1,("Error %d on filldir of %s",rc ,pUnixFindData->FileName));
+	}
 	dput(tmp_dentry);
+	return rc;
 }
 
 int
@@ -1238,8 +1460,10 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 					rc = 0;
 					break;
 				}
-			} else
+			} else {
+				cifsFile->invalidHandle = TRUE;
 				CIFSFindClose(xid, pTcon, cifsFile->netfid);
+			}
 			if(cifsFile->search_resume_name) {
 				kfree(cifsFile->search_resume_name);
 				cifsFile->search_resume_name = NULL;
@@ -1263,6 +1487,8 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 				cifsFile =
 				    (struct cifsFileInfo *) file->private_data;
 				cifsFile->netfid = searchHandle;
+				cifsFile->invalidHandle = FALSE;
+				init_MUTEX(&cifsFile->fh_sem);
 			} else {
 				rc = -ENOMEM;
 				break;
@@ -1347,10 +1573,16 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 						    FileName[0] != '.')
 						|| (pfindData->
 						    FileName[1] != '.'))) {
-						cifs_filldir(&qstring,
+						if(cifs_filldir(&qstring,
 							     pfindData,
 							     file, filldir,
-							     direntry);
+							     direntry)) {
+							/* do not end search if
+								kernel not ready to take
+								remaining entries yet */
+							findParms.EndofSearch = 0;
+							break;
+						}
 						file->f_pos++;
 					}
 				} else {	/* UnixSearch */
@@ -1377,11 +1609,17 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 						    FileName[0] != '.')
 						|| (pfindDataUnix->
 						    FileName[1] != '.'))) {
-						cifs_filldir_unix(&qstring,
+						if(cifs_filldir_unix(&qstring,
 								  pfindDataUnix,
 								  file,
 								  filldir,
-								  direntry);
+								  direntry)) {
+							/* do not end search if
+								kernel not ready to take
+								remaining entries yet */
+							findParms.EndofSearch = 0;
+							break;
+						}
 						file->f_pos++;
 					}
 				}
@@ -1449,6 +1687,11 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 						rc = -ENOMEM;
 						break;
 					}
+					/* Free the memory allocated by previous findfirst 
+					or findnext call - we can not reuse the memory since
+					the resume name may not be same string length */
+					if(cifsFile->search_resume_name)
+						kfree(cifsFile->search_resume_name);
 					cifsFile->search_resume_name = 
 						kmalloc(cifsFile->resume_name_length, GFP_KERNEL);
 					cFYI(1,("Last file: %s with name %d bytes long",
@@ -1479,6 +1722,11 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 						rc = -ENOMEM;
 						break;
 					}
+					/* Free the memory allocated by previous findfirst 
+					or findnext call - we can not reuse the memory since
+					the resume name may not be same string length */
+					if(cifsFile->search_resume_name)
+						kfree(cifsFile->search_resume_name);
 					cifsFile->search_resume_name = 
 						kmalloc(cifsFile->resume_name_length, GFP_KERNEL);
 					cFYI(1,("fnext last file: %s with name %d bytes long",
@@ -1510,11 +1758,17 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 							|| (pfindData->FileName[0] != '.')
 							|| (pfindData->FileName[1] !=
 							    '.'))) {
-							cifs_filldir
+							if(cifs_filldir
 							    (&qstring,
 							     pfindData,
 							     file, filldir,
-							     direntry);
+							     direntry)) {
+							/* do not end search if
+								kernel not ready to take
+								remaining entries yet */
+								findNextParms.EndofSearch = 0;
+								break;
+							}
 							file->f_pos++;
 						}
 					} else {	/* UnixSearch */
@@ -1544,11 +1798,17 @@ cifs_readdir(struct file *file, void *direntry, filldir_t filldir)
 							|| (pfindDataUnix->
 							    FileName[1] !=
 							    '.'))) {
-							cifs_filldir_unix
+							if(cifs_filldir_unix
 							    (&qstring,
 							     pfindDataUnix,
 							     file, filldir,
-							     direntry);
+							     direntry)) {
+								/* do not end search if
+								kernel not ready to take
+								remaining entries yet */
+								findNextParms.EndofSearch = 0;
+								break;
+							}
 							file->f_pos++;
 						}
 					}
