@@ -1,7 +1,7 @@
 /*
  *   fs/cifs/inode.c
  *
- *   Copyright (c) International Business Machines  Corp., 2002
+ *   Copyright (C) International Business Machines  Corp., 2002,2003
  *   Author(s): Steve French (sfrench@us.ibm.com)
  *
  *   This library is free software; you can redistribute it and/or modify
@@ -21,6 +21,7 @@
 #include <linux/fs.h>
 #include <linux/stat.h>
 #include <linux/pagemap.h>
+#include <linux/version.h>
 #include <asm/div64.h>
 #include "cifsfs.h"
 #include "cifspdu.h"
@@ -40,6 +41,10 @@ struct inode * get_cifs_inode(struct super_block * sb)
 		cifsInfo->clientCanCacheAll = FALSE;
 		INIT_LIST_HEAD(&cifsInfo->openFileList);
 		cifsInfo->cifsAttrs = 0x20;   /* default */
+		newinode->i_blksize = CIFS_MAX_MSGSIZE;
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
+		newinode->i_blkbits = 14;  /* 2**14 = CIFS_MAX_MSGSIZE */
+#endif
 		atomic_set(&cifsInfo->inUse, 0);
 		cifsInfo->time = 0;
 	}
@@ -147,8 +152,13 @@ cifs_get_inode_info_unix(struct inode **pinode,
 	and blkbits set in superblock so 2**blkbits and blksize will match */
 /*		inode->i_blksize =
 		    (pTcon->ses->server->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;*/
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 		inode->i_blocks = 
 	                (inode->i_blksize - 1 + findData.NumOfBytes) >> inode->i_blkbits;
+#else
+		inode->i_blocks = 
+	                (inode->i_blksize - 1 + findData.NumOfBytes) >> 14;
+#endif
 
 		if (findData.NumOfBytes < findData.EndOfFile)
 			cFYI(1, ("Server inconsistency Error: it says allocation size less than end of file "));
@@ -291,8 +301,13 @@ cifs_get_inode_info(struct inode **pinode, const unsigned char *search_path,
 		}
 		inode->i_size = le64_to_cpu(pfindData->EndOfFile);
 		pfindData->AllocationSize = le64_to_cpu(pfindData->AllocationSize);
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,15))
 		inode->i_blocks =
 	                (inode->i_blksize - 1 + pfindData->AllocationSize) >> inode->i_blkbits;
+#else
+		inode->i_blocks =
+	                (inode->i_blksize - 1 + pfindData->AllocationSize) >> 14;
+#endif
 
 		cFYI(1,
 		     (" Size %ld and blocks %ld ",
@@ -373,8 +388,9 @@ cifs_unlink(struct inode *inode, struct dentry *direntry)
 				CREATE_NOT_DIR | CREATE_DELETE_ON_CLOSE,
 				&netfid, &oplock, NULL, cifs_sb->local_nls);
 		if(rc==0) {
+			CIFSSMBRenameOpenFile(xid,pTcon,netfid,
+				NULL, cifs_sb->local_nls);
 			CIFSSMBClose(xid, pTcon, netfid);
-			/* BB In the future chain close with the NTCreateX to narrow window */
 			direntry->d_inode->i_nlink--;
 		}
 	} else if (rc == -EACCES) {
@@ -390,8 +406,22 @@ cifs_unlink(struct inode *inode, struct dentry *direntry)
 		}
 		if(rc==0) {
 			rc = CIFSSMBDelFile(xid, pTcon, full_path, cifs_sb->local_nls);
-			if (!rc)
+			if (!rc) {
 				direntry->d_inode->i_nlink--;
+			} else if (rc == -ETXTBSY) {
+				int oplock = FALSE;
+				__u16 netfid;
+
+				rc = CIFSSMBOpen(xid, pTcon, full_path, FILE_OPEN, DELETE,
+                                	CREATE_NOT_DIR | CREATE_DELETE_ON_CLOSE,
+	                                &netfid, &oplock, NULL, cifs_sb->local_nls);
+				if(rc==0) {
+					CIFSSMBRenameOpenFile(xid,pTcon,netfid,NULL,cifs_sb->local_nls);
+					CIFSSMBClose(xid, pTcon, netfid);
+		                        direntry->d_inode->i_nlink--;
+				}
+			/* BB if rc = -ETXTBUSY goto the rename logic BB */
+			}
 		}
 	}
 	cifsInode = CIFS_I(direntry->d_inode);
@@ -429,6 +459,7 @@ cifs_mkdir(struct inode *inode, struct dentry *direntry, int mode)
 	rc = CIFSSMBMkDir(xid, pTcon, full_path, cifs_sb->local_nls);
 	if (rc) {
 		cFYI(1, ("cifs_mkdir returned 0x%x ", rc));
+		d_drop(direntry);
 	} else {
 		inode->i_nlink++;
 		if (pTcon->ses->capabilities & CAP_UNIX)
@@ -531,6 +562,20 @@ cifs_rename(struct inode *source_inode, struct dentry *source_direntry,
 		rc = CIFSSMBRename(xid, pTcon, fromName, toName,
 				   cifs_sb_source->local_nls);
 	}
+
+	if((rc == -EIO)||(rc == -EEXIST)) {
+                int oplock = FALSE;
+                __u16 netfid;
+
+                rc = CIFSSMBOpen(xid, pTcon, fromName, FILE_OPEN, GENERIC_READ,
+                                CREATE_NOT_DIR,
+                                &netfid, &oplock, NULL, cifs_sb_source->local_nls);
+                if(rc==0) {
+                        CIFSSMBRenameOpenFile(xid,pTcon,netfid,
+                                toName, cifs_sb_source->local_nls);
+                        CIFSSMBClose(xid, pTcon, netfid);
+                }
+	}
 	if (fromName)
 		kfree(fromName);
 	if (toName)
@@ -576,12 +621,24 @@ cifs_revalidate(struct dentry *direntry)
 		}            
 	}
 
-	if (cifs_sb->tcon->ses->capabilities & CAP_UNIX)
-		cifs_get_inode_info_unix(&direntry->d_inode, full_path,
+	if (cifs_sb->tcon->ses->capabilities & CAP_UNIX) {
+		rc = cifs_get_inode_info_unix(&direntry->d_inode, full_path,
 					 direntry->d_sb);
-	else
-		cifs_get_inode_info(&direntry->d_inode, full_path, NULL,
+		if(rc) {
+			cFYI(1,("error on getting revalidate info %d",rc));
+/*			if(rc != -ENOENT)
+				rc = 0; */ /* BB should we cache info on certain errors? */
+		}
+	} else {
+		rc = cifs_get_inode_info(&direntry->d_inode, full_path, NULL,
 				    direntry->d_sb);
+		if(rc) {
+			cFYI(1,("error on getting revalidate info %d",rc));
+/*			if(rc != -ENOENT)
+				rc = 0; */  /* BB should we cache info on certain errors? */
+		}
+	}
+	/* should we remap certain errors, access denied?, to zero */
 
 	/* BB if not oplocked, invalidate inode pages if mtime has changed */
 
@@ -604,12 +661,14 @@ void
 cifs_truncate_file(struct inode *inode)
 {				/* BB remove - may not need this function after all BB */
 	int xid;
-	int rc = 0;
+	int rc = -EIO;
+	int found = FALSE;
 	struct cifsFileInfo *open_file = NULL;
 	struct cifs_sb_info *cifs_sb;
 	struct cifsTconInfo *pTcon;
 	struct cifsInodeInfo *cifsInode;
 	struct dentry *dirent;
+	struct list_head * tmp;
 	char *full_path = NULL;   
 
 	xid = GetXid();
@@ -617,6 +676,41 @@ cifs_truncate_file(struct inode *inode)
 	cifs_sb = CIFS_SB(inode->i_sb);
 	pTcon = cifs_sb->tcon;
 
+	/* To avoid spurious oplock breaks from server, in the case
+		of inodes that we already have open, avoid doing path
+		based setting of file size if we can do it by handle.
+		This keeps our caching token (oplock) and avoids
+		timeouts when the local oplock break takes longer to flush
+		writebehind data than the SMB timeout for the SetPathInfo 
+		request would allow */
+	read_lock(&GlobalSMBSeslock); 
+	cifsInode = CIFS_I(inode);
+	list_for_each(tmp, &cifsInode->openFileList) {            
+		open_file = list_entry(tmp,struct cifsFileInfo, flist);
+		/* We check if file is open for writing first */
+		if((open_file->pfile) && (!open_file->invalidHandle) &&
+		   ((open_file->pfile->f_flags & O_RDWR) || 
+			(open_file->pfile->f_flags & O_WRONLY))) {
+			read_unlock(&GlobalSMBSeslock);
+			found = TRUE;
+			rc = CIFSSMBSetFileSize(xid, pTcon, inode->i_size,
+			   open_file->netfid,open_file->pid,FALSE);
+			if(rc == 0) {
+				FreeXid(xid);
+				return;
+			} 
+			/* Do not need reopen and retry on EAGAIN since we will
+				retry by pathname below */
+			if(rc == -EAGAIN)
+				rc = -EHOSTDOWN;
+
+			break;  /* now that we found one valid file handle no
+				sense continuing to loop trying others */
+		}
+	}
+	if(found == FALSE)
+		read_unlock(&GlobalSMBSeslock);
+
 	if (list_empty(&inode->i_dentry)) {
 		cERROR(1,
 		       ("Can not get pathname from empty dentry in inode 0x%p ",
@@ -624,24 +718,13 @@ cifs_truncate_file(struct inode *inode)
 		FreeXid(xid);
 		return;
 	}
+
 	dirent = list_entry(inode->i_dentry.next, struct dentry, d_alias);
 	if (dirent) {
 		full_path = build_path_from_dentry(dirent);
 		rc = CIFSSMBSetEOF(xid, pTcon, full_path, inode->i_size,FALSE,
 				   cifs_sb->local_nls);
 		cFYI(1,(" SetEOF (truncate) rc = %d",rc));
-		if(rc == -ETXTBSY) {        
-			cifsInode = CIFS_I(inode);
-			if(!list_empty(&(cifsInode->openFileList))) {            
-				open_file = list_entry(cifsInode->openFileList.next,
-					struct cifsFileInfo, flist);           
-            /* We could check if file is open for writing first */
-				 rc = CIFSSMBSetFileSize(xid, pTcon, inode->i_size,
-					open_file->netfid,open_file->pid,FALSE);
-			} else {
-				  cFYI(1,(" No open files to get file handle from"));
-			}
-		}
 		if (!rc)
 			CIFSSMBSetEOF(xid,pTcon,full_path,inode->i_size,TRUE,cifs_sb->local_nls);
            /* allocation size setting seems optional so ignore return code */
@@ -669,7 +752,11 @@ static int cifs_trunc_page(struct address_space *mapping, loff_t from)
         flush_dcache_page(page);
         kunmap_atomic(kaddr, KM_USER0);
         set_page_dirty(page);
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,20))
         unlock_page(page);
+#else
+	UnlockPage(page);
+#endif
         page_cache_release(page);
         return rc;
 }
@@ -682,6 +769,7 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	struct cifsTconInfo *pTcon;
 	char *full_path = NULL;
 	int rc = -EACCES;
+	int found = FALSE;
 	struct cifsFileInfo *open_file = NULL;
 	FILE_BASIC_INFO time_buf;
 	int set_time = FALSE;
@@ -689,6 +777,7 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	__u64 uid = 0xFFFFFFFFFFFFFFFFULL;
 	__u64 gid = 0xFFFFFFFFFFFFFFFFULL;
 	struct cifsInodeInfo *cifsInode;
+	struct list_head * tmp;
 
 	xid = GetXid();
 
@@ -706,24 +795,54 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	cFYI(1, (" Changing attributes 0x%x", attrs->ia_valid));
 
 	if (attrs->ia_valid & ATTR_SIZE) {
-		rc = CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size,FALSE,
-				   cifs_sb->local_nls);
-		cFYI(1,(" SetEOF (setattrs) rc = %d",rc));
+		read_lock(&GlobalSMBSeslock);
+		/* To avoid spurious oplock breaks from server, in the case
+			of inodes that we already have open, avoid doing path
+			based setting of file size if we can do it by handle.
+			This keeps our caching token (oplock) and avoids
+			timeouts when the local oplock break takes longer to flush
+			writebehind data than the SMB timeout for the SetPathInfo
+			request would allow */
+		list_for_each(tmp, &cifsInode->openFileList) {
+			open_file = list_entry(tmp,struct cifsFileInfo, flist);
+			/* We check if file is open for writing first */
+			if((open_file->pfile) &&
+			   ((open_file->pfile->f_flags & O_RDWR) ||
+			    (open_file->pfile->f_flags & O_WRONLY))) {
+				if(open_file->invalidHandle == FALSE) {
+					/* we found a valid, writeable network file
+					handle to use to try to set the file size */
+					__u16 nfid = open_file->netfid;
+					__u32 npid = open_file->pid;
+					read_unlock(&GlobalSMBSeslock);
+					found = TRUE;
+					rc = CIFSSMBSetFileSize(xid, pTcon, attrs->ia_size,
+						nfid,npid,FALSE);
+					cFYI(1,("SetFileSize by handle (setattrs) rc = %d",rc));
+				/* Do not need reopen and retry on EAGAIN since we will
+					retry by pathname below */
 
-		if(rc == -ETXTBSY) {
-			if(!list_empty(&(cifsInode->openFileList))) {            
-				open_file = list_entry(cifsInode->openFileList.next, 
-					   struct cifsFileInfo, flist);           
-    /* We could check if file is open for writing first */
-				rc = CIFSSMBSetFileSize(xid, pTcon, attrs->ia_size,
-					   open_file->netfid,open_file->pid,FALSE);           
-			} else {
-				cFYI(1,(" No open files to get file handle from"));
+					break;  /* now that we found one valid file handle no
+						sense continuing to loop trying others */
+				}
 			}
 		}
+		if(found == FALSE)
+			read_unlock(&GlobalSMBSeslock);
+
+		if(rc != 0) {
+			/* Set file size by pathname rather than by handle either
+			because no valid, writeable file handle for it was found or
+			because there was an error setting it by handle */
+			rc = CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size,FALSE,
+				cifs_sb->local_nls);
+			cFYI(1,(" SetEOF by path (setattrs) rc = %d",rc));
+		}
+
         /*  For Allocation Size - do not need to call the following
             it did not hurt if it fails but why bother */
 	/*	CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size, TRUE, cifs_sb->local_nls);*/
+
 		if (rc == 0) {
 			rc = vmtruncate(direntry->d_inode, attrs->ia_size);
 			cifs_trunc_page(direntry->d_inode->i_mapping, direntry->d_inode->i_size);
@@ -791,6 +910,8 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 		/* BB what if setting one attribute fails  
 			(such as size) but time setting works */
 		time_buf.CreationTime = 0;	/* do not change */
+		/* In the future we should experiment - try setting timestamps
+			 via Handle (SetFileInfo) instead of by path */
 		rc = CIFSSMBSetTimes(xid, pTcon, full_path, &time_buf,
 				cifs_sb->local_nls);
 	}
